// httpRequest.js
// ------------------------------------------------------------------
//
// created: Tue Jun  7 15:48:13 2022
// last saved: <2022-June-08 19:24:17>

/* jshint esversion:9, node:true, strict:implied */
/* global process, console, Buffer */

const http = require('http'),
      https = require('https'),
      url = require('url');
//const {PassThrough} = require('stream');
//const fs = require('fs');
const noBodyMethods = ['GET', 'HEAD', 'DELETE', 'OPTIONS'];

const keysToLowerCase = (obj) =>
  Object.keys(obj).reduce( (a, key) => (a[key.toLowerCase()] = obj[key], a), {});

const textHandler = (cb) => (res) => {
        res.setEncoding('utf8');
        let payload = '';
        res.on('data', chunk => payload += chunk);
        res.on('end', () => {
          //let r = {payload:JSON.parse(payload), statusCode:res.statusCode, headers:res.headers};
          //let r = JSON.parse(payload);
          let e = (res.complete) ? null : new Error('incomplete response');
          cb(e, res, payload);
        });
        res.on('error', e => cb(e) );
      };

const binHandler = (cb) => (res) => {
        res.setEncoding('binary');
        let data = [];
        res.on('data', chunk => data.push(Buffer.from(chunk, 'binary')));
        res.on('end', () => {
          data = Buffer.concat(data);
          let r = {payload:data, statusCode:res.statusCode, headers:res.headers};
          let e = (res.complete) ? null : new Error('incomplete response');
          cb(e, r);
        });
        res.on('error', e => cb(e));
      };

async function httpRequest(opts, cb) {
  // eg, opts = {
  //       url: 'https://foo-bar.bam/blah',
  //       method: 'post',
  //       body : querystring.stringify({
  //         grant_type : 'urn:ietf:params:oauth:grant-type:jwt-bearer',
  //         assertion: token
  //       }),
  //       headers : {
  //         'content-type': 'application/x-www-form-urlencoded',
  //         accept: 'application/json'
  //       }
  //     };

  //console.log('%s %s', opts.method.toUpperCase(), opts.url);
  let parsed = url.parse(opts.url),
      options = {
          host: parsed.host,
          path: parsed.path,
          method : opts.method,
          headers : keysToLowerCase(opts.headers)
      },
      expectBinary = options.headers.accept && (opts.headers.accept == 'application/octet-stream' ||
                                                opts.headers.accept == '*/*'), // the export API requires this
      dataHandler = expectBinary ? binHandler(cb) : textHandler(cb),
      boundary = '';
  if (parsed.host.indexOf(':')>=0) {
    let parts = parsed.host.split(':');
    options.host = parts[0];
    options.port = parts[1];
  }
  if (noBodyMethods.indexOf(opts.method) == -1) {
    if (opts.formData) {
      boundary = 'x'.repeat(12) + '-' + String(Math.random(36)).slice(2);
      options.headers['content-type'] = `multipart/form-data; boundary=${boundary}`;
    }
  }

  let request = ((parsed.protocol == 'https:') ? https : http).request(options, dataHandler);

  if (noBodyMethods.indexOf(opts.method) >= 0) {
    request.end();
  }
  else {
    if (opts.body) {
      request.write(opts.body);
      request.end();
    }
    else if (opts.formData) {
      // multipart/form-data
      // handle form items, some of them might be streams.
      let keys = Object.keys(opts.formData);
      for (let i = 0; i<keys.length; i++) {
        let key = keys[i];
        let item = opts.formData[key];
        if (item.constructor.name == 'ReadStream') {
          request.write(`--${boundary}\r\n`);
          request.write(`Content-Disposition: form-data; name="${key}";\r\n`);
          request.write("Content-Type:application/octet-stream\r\n\r\n");
          await new Promise((resolve, reject) =>
            item.pipe(request, {end:false}) .on('unpipe', () => resolve() ) );
          request.write(`\r\n`);
        }
        else if (item.constructor.name == 'String') {
          request.write(`--${boundary}\r\n`);
          request.write(`Content-Disposition: form-data; name="${key}";\r\n\r\n${item}\r\n`);
        }
      }
      request.write(`--${boundary}--\r\n`);
      request.end();
    }
  }

  // If POST/PATCH/PUT, and neither body nor formData is specified, then the
  // request stream stays open for writing/piping by the caller.

  return request; // a writeable stream
}

let request = {
      post: function (options, callback) {
        let localOptions = {...options};
        localOptions.method = 'POST';
        return httpRequest(localOptions, callback);
      },
      put : function (options, callback) {
        let localOptions = {...options};
        localOptions.method = 'PUT';
        return httpRequest(localOptions, callback);
      },
      del: function (options, callback) {
        let localOptions = {...options};
        localOptions.method = 'DELETE';
        return httpRequest(localOptions, callback);
      },
      get: function (options, callback) {
        let localOptions = {...options};
        localOptions.method = 'GET';
        return httpRequest(localOptions, callback);
      }
};

module.exports = request;
